@@ -1,12 +1,64 @@
 #include "cairoDevice.h"
 
+// All measurements within this file are made in points (unless otherwise stated)
+#define POINTS_PER_INCH 72
 #define CURSOR		GDK_CROSSHAIR		/* Default cursor */
 #define MM_PER_INCH	25.4			/* mm -> inch conversion */
 
 static PangoStyle  slant[]  = { PANGO_STYLE_NORMAL, PANGO_STYLE_OBLIQUE };
 static PangoWeight weight[] = { PANGO_WEIGHT_NORMAL, PANGO_WEIGHT_BOLD };
 
-CairoDesc *createCairoDesc() {
+// Resolution is measured in dpi (dots per inch)
+double gResolutionX = 72;
+double gResolutionY = 72;
+
+// Sets resolution independently for X and Y directions 
+void
+setResolutionXY(double resX, double resY) {
+  //printf("Resolution was (%f,%f)\n", gResolutionX, gResolutionY);
+  gResolutionX = resX;
+  gResolutionY = resY;
+  //printf("Resolution is now (%f,%f)\n", gResolutionX, gResolutionY);
+}
+
+// Sets resolution assuming X and Y are identical
+void
+setResolution(double res) {
+  setResolutionXY(res,res);
+}
+
+// A function exposed to R allowing it to change the resolution
+SEXP 
+Cairo_Set_Resolution(SEXP res) {
+  double resolution = asReal(res);
+  setResolution(resolution);
+  return ScalarReal(resolution);
+}
+
+// Obtains X and Y resolutions of screen surface and updates global variables
+void
+setResolutionFromScreen() {
+  double width, widthMM;
+  double height, heightMM;  
+  double resX, resY;
+
+  // X direction
+  width = gdk_screen_width();
+  widthMM = gdk_screen_width_mm();
+  resX = width/widthMM * MM_PER_INCH;
+
+  // Y direction
+  height = gdk_screen_height();
+  heightMM = gdk_screen_height_mm();
+  resY = height/heightMM * MM_PER_INCH;
+
+  // Update global variables
+  setResolutionXY(resX, resY);
+}
+
+
+
+CairoDesc* createCairoDesc() {
   return(g_new0(CairoDesc, 1));
 }
 void freeCairoDesc(pDevDesc dd) {
@@ -24,9 +76,11 @@ void freeCairoDesc(pDevDesc dd) {
   
   if(cd->window)
     gtk_widget_destroy(cd->window);
-  
+
   if(cd->cr) {
-    cairo_show_page(cd->cr);
+    if (!cd->cr_custom)
+      cairo_show_page(cd->cr);
+    else cairo_restore(cd->cr);
     cairo_destroy(cd->cr);
   }
   
@@ -54,22 +108,6 @@ static void activateDevice(pDevDesc dev)
   }
 }
 
-/** need the pixel width/height in inches so we can convert 
-    from R's inches to our pixels */
-static double pixelWidth(void)
-{
-  double width, widthMM;
-  width = gdk_screen_width();
-  widthMM = gdk_screen_width_mm();
-  return ((double)widthMM / (double)width) / MM_PER_INCH;
-}
-static double pixelHeight(void)
-{
-  double height, heightMM;
-  height = gdk_screen_height();
-  heightMM = gdk_screen_height_mm();
-  return ((double)heightMM / (double)height) / MM_PER_INCH;
-}
 
 static PangoContext *getPangoContext(CairoDesc *cd)
 {
@@ -97,6 +135,7 @@ static gboolean initDevice(pDevDesc dd)
   cd = (CairoDesc *) dd->deviceSpecific;
 	
   Cairo_Size(&left, &right, &bottom, &top, dd);
+  //printf("Cairo_Size returned (%f %f %f %f)\n", left, right, bottom, top);
   /*g_debug("%f %f %f %f" , left, right, bottom, top);*/
   dd->left = left;
   cd->width = dd->right = right;
@@ -111,26 +150,39 @@ static gboolean initDevice(pDevDesc dd)
   }
   
   if(cd->cr) {
-    cairo_show_page(cd->cr);
-    cairo_destroy(cd->cr);
+    if (!cd->cr_custom) {
+      cairo_show_page(cd->cr);
+      cairo_destroy(cd->cr);
+    } else cairo_restore(cd->cr);
   }
   if(cd->pixmap && cd->drawing)
     g_object_unref(cd->pixmap);
   
   /* create pixmap and cairo context for drawing, save default state */
   if(right > 0 && bottom > 0) {
-    if (cd->drawing)
-      cd->pixmap = gdk_pixmap_new(cd->drawing->window, right, bottom, -1);
-    if (cd->surface)
+    if (cd->drawing) {
+      cd->pixmap = gdk_pixmap_new(cd->drawing->window, 
+				  right*gResolutionX/POINTS_PER_INCH, 
+				  bottom*gResolutionY/POINTS_PER_INCH, -1);
+    }
+    if (cd->surface) {
       cd->cr = cairo_create(cd->surface);
+    }
+    else if (cd->cr_custom) {
+      cd->cr = cd->cr_custom;
+    }
     else cd->cr = gdk_cairo_create(cd->pixmap);
   //cairo_set_antialias(cd->cr, CAIRO_ANTIALIAS_NONE);
     //cd->pango = pango_cairo_font_map_create_context(
     //				PANGO_CAIRO_FONT_MAP(pango_cairo_font_map_get_default()));
     //pango_cairo_context_set_resolution(cd->pango, 84);
-    cairo_save(cd->cr);
     /* sync the size with the device */
   }
+
+  // Change the CTM (current transformation matrix) to scale for different resolutions
+  cairo_scale(cd->cr, gResolutionX/POINTS_PER_INCH,gResolutionY/POINTS_PER_INCH);
+  //printf("  scaling by %f, %f\n", gResolutionX/POINTS_PER_INCH,gResolutionY/POINTS_PER_INCH);
+    cairo_save(cd->cr);
   
   return FALSE;
 }
@@ -164,8 +216,13 @@ static gint expose_event(GtkWidget *widget, GdkEventExpose *event, pDevDesc dd)
   g_return_val_if_fail(cd != NULL, FALSE);
   g_return_val_if_fail(GTK_IS_DRAWING_AREA(cd->drawing), FALSE);
 
-  if(cd->width != cd->drawing->allocation.width || cd->height != cd->drawing->allocation.height) 
+  gint current_width = ((double)cd->drawing->allocation.width/gResolutionX*POINTS_PER_INCH);
+  gint current_height = ((double)cd->drawing->allocation.height/gResolutionY*POINTS_PER_INCH);
+
+  if(cd->width != current_width
+     || cd->height != current_height) {
     resize(dd); 
+  }
   else if(cd->pixmap) {
     gdk_draw_drawable(cd->drawing->window, cd->drawing->style->bg_gc[GTK_STATE_NORMAL], 
                       cd->pixmap,
@@ -179,6 +236,7 @@ static gint expose_event(GtkWidget *widget, GdkEventExpose *event, pDevDesc dd)
   return FALSE;
 }
 
+#if R_GE_version < 7
 static void event_finish(pDevDesc dd)
 {
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
@@ -197,6 +255,7 @@ static void CairoEvent_onExit(pDevDesc dd)
 {
   event_finish(dd);
 }
+#endif
 
 /* converts the Gdk buttons to the button masks used by R's gevent.c */
 #define R_BUTTON(button) pow(2, (button) - 1)
@@ -204,47 +263,65 @@ static void CairoEvent_onExit(pDevDesc dd)
 
 static gboolean button_press_event(GtkWidget *widget, GdkEventButton *event, pDevDesc dd)
 {
-  CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->gettingEvent) {
+#if R_GE_version < 7
+    CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
     cd->event->result = doMouseEvent(cd->event->rho, dd, meMouseDown,
                                      R_BUTTON(event->button), event->x, event->y);
     event_maybe_finish(dd);
+#else
+    doMouseEvent(dd, meMouseDown, R_BUTTON(event->button), event->x, event->y);
+#endif
+
   }
   return(FALSE);
 }
 
 static gboolean button_release_event(GtkWidget *widget, GdkEventButton *event, pDevDesc dd)
 {
-  CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->gettingEvent) {
+#if R_GE_version < 7
+    CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
     cd->event->result = doMouseEvent(cd->event->rho, dd, meMouseUp,
                                      R_BUTTON(event->button), event->x, event->y);
     event_maybe_finish(dd);
+#else
+    doMouseEvent(dd, meMouseUp, R_BUTTON(event->button), event->x, event->y);
+#endif
   }
   return(FALSE);
 }
 
 static gboolean motion_notify_event(GtkWidget *widget, GdkEventMotion *event, pDevDesc dd)
 {
-  CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->gettingEvent) {
+#if R_GE_version < 7
+    CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
     cd->event->result = doMouseEvent(cd->event->rho, dd, meMouseMove,
                                      R_BUTTONS(event->state), event->x, event->y);
     event_maybe_finish(dd);
+#else
+    doMouseEvent(dd, meMouseMove, R_BUTTONS(event->state), event->x, event->y);
+#endif
   }
   return(FALSE);
 }
 
 static gboolean key_press_event(GtkWidget *widget, GdkEventKey *event, pDevDesc dd)
 {
-  CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
   if (dd->gettingEvent) {
+#if R_GE_version < 7
+    CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
     cd->event->result = doKeybd(cd->event->rho, dd, -1, gdk_keyval_name(event->keyval));
     event_maybe_finish(dd);
+#else
+    doKeybd(dd, -1, gdk_keyval_name(event->keyval));
+#endif
   }
   return(FALSE);
 }
 
+#if R_GE_version < 7
 static SEXP Cairo_GetEvent(SEXP rho, const char *prompt)
 {
   pGEDevDesc dd = GEcurrentDevice();
@@ -276,6 +353,25 @@ static SEXP Cairo_GetEvent(SEXP rho, const char *prompt)
   g_free(event);
   return result;
 }
+#else
+static void Cairo_EventHelper(pDevDesc dd, int code)
+{
+  if (code == 1) {
+    if (isEnvironment(dd->eventEnv)) {
+      SEXP prompt = findVar(install("prompt"), dd->eventEnv);
+      if (length(prompt) == 1) {
+        const char* cprompt = CHAR(asChar(prompt));
+        R_WriteConsole(cprompt, strlen(cprompt));
+        R_WriteConsole("\n", 1);
+        R_FlushConsole();
+      }
+    }
+  } else if (code == 2) 
+    R_gtk_eventHandler(NULL);
+  
+  return;
+}
+#endif
 
 static void kill_cairo(pDevDesc dd)
 {
@@ -333,6 +429,19 @@ static Rboolean Cairo_OpenOffscreen(pDevDesc dd, CairoDesc *cd, GdkDrawable *dra
   initDevice(dd);		 
   return(TRUE);
 }
+
+static Rboolean Cairo_OpenCustom(pDevDesc dd, CairoDesc *cd, double w, double h,
+                                 cairo_t *cr)
+{
+  dd->deviceSpecific = cd;
+  cd->cr_custom = cairo_reference(cr);
+  cd->width = w;
+  cd->height = h;
+  // initialize
+  initDevice(dd);		 
+  return(TRUE);
+}
+
 static Rboolean Cairo_Open(pDevDesc dd, CairoDesc *cd,	double w, double h, 
                            const gchar **surface_info)
 {	
@@ -342,13 +451,15 @@ static Rboolean Cairo_Open(pDevDesc dd, CairoDesc *cd,	double w, double h,
     cairo_surface_t *surface;
     double width, height;
     if (!strcmp(surface_info[0], "png")) {
-      width = w / pixelWidth();
-      height = h / pixelHeight();
+      width = w * gResolutionX/POINTS_PER_INCH;
+      height = h * gResolutionY/POINTS_PER_INCH;
       surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
       cd->filename = g_strdup(surface_info[1]);
     } else {
-      width = w*72.0;
-      height = h*72.0;
+      
+      setResolution(POINTS_PER_INCH);
+      width = w;
+      height = h;
 #ifdef CAIRO_HAS_PDF_SURFACE
       if (!strcmp(surface_info[0], "pdf"))
         surface = cairo_pdf_surface_create(surface_info[1], width, height);
@@ -369,16 +480,20 @@ static Rboolean Cairo_Open(pDevDesc dd, CairoDesc *cd,	double w, double h,
               return(FALSE);
             }
     }
-    cd->width = width;
-    cd->height = height;
+    cd->width = w;
+    cd->height = h;
     cd->surface = surface;
     return(TRUE);
   }
 
+  // Set resolution
+  setResolutionFromScreen(); 
+  
   cd->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_resizable(GTK_WINDOW(cd->window), TRUE);
   gtk_window_set_default_size(GTK_WINDOW(cd->window), 
-                              w / pixelWidth(), h / pixelHeight());
+                              w * gResolutionX/POINTS_PER_INCH, 
+			      h * gResolutionY/POINTS_PER_INCH);
 
   /* create drawingarea */
   cd->drawing = gtk_drawing_area_new();
@@ -398,42 +513,35 @@ static Rboolean Cairo_Open(pDevDesc dd, CairoDesc *cd,	double w, double h,
                    G_CALLBACK(key_press_event), dd);
   
   gtk_widget_show_all(cd->window);
-  
+
   return(TRUE);
 }
 
 static PangoFontDescription *getFont(CairoDesc *cd, const pGEcontext gc)
 {
-  PangoFontDescription *fontdesc, *basedesc;
-  gint size, face = gc->fontface;
-    
-  if (face < 1 || face > 5)
-    face = 1;
+  PangoFontDescription *fontdesc;
+  gint face = gc->fontface;
+  double size = gc->cex * gc->ps;
+
+  if (face < 1 || face > 5) face = 1;
 
-  size = gc->cex * gc->ps + 0.5;
-	
-  basedesc = getBaseFont(cd);
   fontdesc = pango_font_description_new();
-  if (face == SYMBOL_FONTFACE) {
+  if (face == 5)
     pango_font_description_set_family(fontdesc, "symbol");
-  } else {
-    if (strlen(gc->fontfamily))
-      pango_font_description_set_family(fontdesc, gc->fontfamily);
-    pango_font_description_set_weight(fontdesc, weight[(face-1)%2]);
-    pango_font_description_set_style(fontdesc, slant[((face-1)/2)%2]);
+  else {
+    char *fm = gc->fontfamily;
+    /*if(!strcmp(fm, "mono")) fm = "courier";
+    else if(!strcmp(fm, "serif")) fm = "times";
+    else if(!strcmp(fm, "sans")) fm = "helvetica";*/
+    pango_font_description_set_family(fontdesc, fm[0] ? fm : "Verdana");
+    if(face == 2 || face == 4)
+      pango_font_description_set_weight(fontdesc, PANGO_WEIGHT_BOLD);
+    if(face == 3 || face == 4)
+      pango_font_description_set_style(fontdesc, PANGO_STYLE_OBLIQUE);
   }
   pango_font_description_set_size(fontdesc, PANGO_SCALE * size);
-  pango_font_description_merge(fontdesc, basedesc, FALSE);
-	
-  if (!loadFont(fontdesc, cd)) {
-    char *strdesc = pango_font_description_to_string(fontdesc);
-    fprintf(stderr, "Could not find font '%s', falling back to base font!\n", strdesc);
-    g_free(strdesc);
-    pango_font_description_free(fontdesc);
-    fontdesc = basedesc;
-  } else pango_font_description_free(basedesc);
-	
-  return(fontdesc);
+
+  return fontdesc;
 }
 
 static PangoLayout *layoutText(PangoFontDescription *desc, const char *str, 
@@ -443,9 +551,10 @@ static PangoLayout *layoutText(PangoFontDescription *desc, const char *str,
 	
   //pango_cairo_update_context(cd->cr, cd->pango);
   //layout = pango_layout_new(cd->pango);
-  if (cd->drawing)
+  /*if (cd->drawing)
     layout = gtk_widget_create_pango_layout(cd->drawing, NULL);
-  else layout = pango_layout_new(gdk_pango_context_get());
+    else layout = pango_layout_new(gdk_pango_context_get());*/
+  layout = pango_cairo_create_layout(cd->cr);
   pango_layout_set_font_description(layout, desc);
   pango_layout_set_text(layout, str, -1);
   return(layout);
@@ -455,26 +564,40 @@ static void
 text_extents(PangoFontDescription *desc, CairoDesc *cd, const pGEcontext gc, 
              const gchar *text,
 	     gint *lbearing, gint *rbearing, 
-	     gint *width, gint *ascent, gint *descent)
+	     gint *width, gint *ascent, gint *descent, gboolean ink)
 {
   PangoLayout *layout;
-  PangoRectangle rect;
+  PangoRectangle rect, irect;
 	
   layout = layoutText(desc, text, cd);
   
-  pango_layout_line_get_pixel_extents(pango_layout_get_line(layout, 0), NULL, &rect);
+  pango_layout_line_get_pixel_extents(pango_layout_get_line(layout, 0), &irect,
+                                      &rect);
 
-  if(ascent)
-    *ascent = PANGO_ASCENT(rect);
-  if(descent)
-    *descent = PANGO_DESCENT(rect);
-  if(width)
-    *width = rect.width;
-  if(lbearing)
-    *lbearing = PANGO_LBEARING(rect);
-  if(rbearing)
-    *rbearing = PANGO_RBEARING(rect);
-	
+  if (!ink) {
+    if(ascent)
+      *ascent = PANGO_ASCENT(rect);
+    if(descent)
+      *descent = PANGO_DESCENT(rect);
+    if(width)
+      *width = rect.width;
+    if(lbearing)
+      *lbearing = PANGO_LBEARING(rect);
+    if(rbearing)
+      *rbearing = PANGO_RBEARING(rect);
+  } else {
+    if(ascent)
+      *ascent = PANGO_ASCENT(irect);
+    if(descent)
+      *descent = PANGO_DESCENT(irect);
+    if(width)
+      *width = irect.width;
+    if(lbearing)
+      *lbearing = PANGO_LBEARING(irect);
+    if(rbearing)
+      *rbearing = PANGO_RBEARING(irect);
+  }
+  
   g_object_unref(layout);
 }
 
@@ -573,7 +696,7 @@ static double Cairo_StrWidth(const char *str, const pGEcontext gc, pDevDesc dd)
   CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
 	
   PangoFontDescription *desc = getFont(cd, gc);
-  text_extents(desc, cd, gc, str, NULL, NULL, &width, NULL, NULL);
+  text_extents(desc, cd, gc, str, NULL, NULL, &width, NULL, NULL, FALSE);
   pango_font_description_free(desc);
 	
   return (double) width;
@@ -592,14 +715,16 @@ static void Cairo_MetricInfo(int c, const pGEcontext gc,
   if (!c) 
     font_metrics(desc, cd, &iwidth, &iascent, &idescent);
   else {
-      if(c < 0) {c = -c; Unicode = 1;} 
-      
-      if(Unicode || c >= 128)
-	  Rf_ucstoutf8(text, c);
-      else
-	  g_snprintf(text, 2, "%c", (gchar) c);
-      text_extents(desc, cd, gc, text, NULL, NULL, &iwidth, &iascent, &idescent);
-  }
+    /*if (!c) c = 77;*/
+    if(c < 0) {c = -c; Unicode = 1;} 
+  
+  if(Unicode || c >= 128)
+    Rf_ucstoutf8(text, c);
+  else
+    g_snprintf(text, 2, "%c", (gchar) c);
+  text_extents(desc, cd, gc, text, NULL, NULL, &iwidth, &iascent, &idescent,
+               TRUE);
+      }
 	
   *ascent = iascent;
   *descent = idescent;
@@ -639,11 +764,16 @@ static void Cairo_Size(double *left, double *right, double *bottom, double *top,
      the new backing pixmap is not created until later */
   if (cd->drawing) {
     if (GTK_WIDGET_MAPPED(cd->drawing)) {
-      width = cd->drawing->allocation.width;
-      height = cd->drawing->allocation.height;
-    } else width = height = 1000; // hack to get R to draw before first exposure
-  } else if (GDK_IS_DRAWABLE(cd->pixmap))
+      width = (double)cd->drawing->allocation.width / gResolutionX * POINTS_PER_INCH ;
+      height = (double)cd->drawing->allocation.height / gResolutionY * POINTS_PER_INCH;
+    } else {
+      width = height = 1000; // hack to get R to draw before first exposure
+    }
+  } else if (GDK_IS_DRAWABLE(cd->pixmap)) {
     gdk_drawable_get_size(cd->pixmap, &width, &height);
+    width = (double)width / gResolutionX * POINTS_PER_INCH;
+    height = (double)height / gResolutionY * POINTS_PER_INCH;
+  }
 
   *left = 0.0;
   *right = (gdouble)width;
@@ -830,26 +960,140 @@ static void Cairo_Polygon(int n, double *x, double *y,
   cairo_restore(cd->cr);
 }
 
+/* Adapted from R's modules/X11/cairoX11.c */
+static void Cairo_Raster(unsigned int *raster, int w, int h,
+                         double x, double y, 
+                         double width, double height,
+                         double rot, 
+                         Rboolean interpolate,
+                         const pGEcontext gc, pDevDesc dd)
+{
+  char *vmax = vmaxget();
+  int i;
+  cairo_surface_t *image;
+  unsigned char *imageData;
+  CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
+  
+  imageData = (unsigned char *) R_alloc(4*w*h, sizeof(unsigned char));
+  /* The R ABGR needs to be converted to a Cairo ARGB 
+   * AND values need to by premultiplied by alpha 
+   */
+  for (i=0; i<w*h; i++) {
+    int alpha = R_ALPHA(raster[i]);
+    imageData[i*4 + 3] = alpha;
+    if (alpha < 255) {
+      imageData[i*4 + 2] = R_RED(raster[i]) * alpha / 255;
+      imageData[i*4 + 1] = R_GREEN(raster[i]) * alpha / 255;
+      imageData[i*4 + 0] = R_BLUE(raster[i]) * alpha / 255;
+    } else {
+      imageData[i*4 + 2] = R_RED(raster[i]);
+      imageData[i*4 + 1] = R_GREEN(raster[i]);
+      imageData[i*4 + 0] = R_BLUE(raster[i]);
+    }
+  }
+  image = cairo_image_surface_create_for_data(imageData, 
+                                              CAIRO_FORMAT_ARGB32,
+                                              w, h, 
+                                              4*w);
+
+  cairo_save(cd->cr);
+
+  cairo_translate(cd->cr, x, y);
+  cairo_rotate(cd->cr, -rot*M_PI/180);
+  cairo_scale(cd->cr, width/w, height/h);
+  /* Flip vertical first */
+  cairo_translate(cd->cr, 0, h/2.0);
+  cairo_scale(cd->cr, 1, -1);
+  cairo_translate(cd->cr, 0, -h/2.0);
+
+  cairo_set_source_surface(cd->cr, image, 0, 0);
+
+  /* Use nearest-neighbour filter so that a scaled up image
+   * is "blocky";  alternative is some sort of linear
+   * interpolation, which gives nasty edge-effects
+   */
+  if (!interpolate) {
+    cairo_pattern_set_filter(cairo_get_source(cd->cr), 
+                             CAIRO_FILTER_NEAREST);
+  }
+
+  cairo_paint(cd->cr); 
+
+  cairo_restore(cd->cr);
+  cairo_surface_destroy(image);
+
+  vmaxset(vmax);
+}
+
+static SEXP Cairo_Cap(pDevDesc dd)
+{
+  int i, j, k, nbytes, width, height, size, stride;
+  CairoDesc *cd = (CairoDesc *) dd->deviceSpecific;
+  unsigned char *screenData;
+  SEXP dim, raster = R_NilValue;
+  unsigned int *rint;
+  GdkPixbuf *pixbuf = NULL;
+  
+  if (!cd->pixmap)
+    return raster;
+
+  pixbuf =
+    gdk_pixbuf_get_from_drawable(NULL, cd->pixmap, NULL, 0, 0, 0, 0, -1, -1);
+
+  stride = gdk_pixbuf_get_rowstride(pixbuf);
+  width = gdk_pixbuf_get_width(pixbuf);
+  height = gdk_pixbuf_get_height(pixbuf);
+  screenData = gdk_pixbuf_get_pixels(pixbuf);
+  
+  /* Should not happen */
+  if (gdk_pixbuf_get_colorspace(pixbuf) != GDK_COLORSPACE_RGB ||
+      gdk_pixbuf_get_bits_per_sample(pixbuf) != 8 ||
+      gdk_pixbuf_get_has_alpha(pixbuf)) 
+    return raster;
+  
+  size = width*height;
+  nbytes = stride*height;
+
+  PROTECT(raster = allocVector(INTSXP, size));
+
+  /* Copy each byte of screen to an R matrix. 
+   * The Cairo RGB24 needs to be converted to an R ABGR32 */
+  rint = (unsigned int *) INTEGER(raster);
+  for (i = 0, k = 0; i < nbytes; i += stride)
+    for (j = i; j < i + width*3; j += 3)
+      rint[k++] = 255<<24 | ((screenData[j])<<16 | 
+                             (screenData[j + 1])<<8 |
+                             (screenData[j + 2]));
+  
+  PROTECT(dim = allocVector(INTSXP, 2));
+  INTEGER(dim)[0] = height;
+  INTEGER(dim)[1] = width;
+  setAttrib(raster, R_DimSymbol, dim);
+
+  UNPROTECT(2);
+  return raster;
+}
+
 static void drawText(double x, double y, const char *str, 
 		     double rot, double hadj, CairoDesc *cd, const pGEcontext gc)
 {
   PangoLayout *layout;
   gint ascent, lbearing;
   cairo_t *cr = cd->cr;
-	
+
   PangoFontDescription *desc = getFont(cd, gc);
 
-  // x and y seem a little off, correcting...
-  text_extents(desc, cd, gc, str, &lbearing, NULL, NULL, &ascent, NULL);
-	
+  layout = layoutText(desc, str, cd);
+  
+  text_extents(desc, cd, gc, str, &lbearing, NULL, NULL, &ascent, NULL, FALSE);
+
   cairo_move_to(cr, x, y);
   cairo_rotate(cr, -1*rot);
   cairo_rel_move_to(cr, -lbearing, -ascent);
   setColor(cr, gc->col);
-	
-  layout = layoutText(desc, str, cd);
+
   pango_cairo_show_layout(cr, layout);
-	
+  
   g_object_unref(layout);
   pango_font_description_free(desc);
 }
@@ -980,10 +1224,16 @@ configureCairoDevice(pDevDesc dd, CairoDesc *cd, double width, double height, do
   dd->line = Cairo_Line;
   dd->polyline = Cairo_Polyline;
   dd->polygon = Cairo_Polygon;
+  dd->raster = Cairo_Raster;
+  dd->cap = Cairo_Cap;
   dd->locator = Cairo_Locator;
   dd->mode = Cairo_Mode;
   dd->metricInfo = Cairo_MetricInfo;
+#if R_GE_version < 7
   dd->getEvent = Cairo_GetEvent;
+#else
+  dd->eventHelper = Cairo_EventHelper;
+#endif
   dd->hasTextUTF8 = TRUE;
   dd->wantSymbolUTF8 = TRUE;
   dd->strWidthUTF8 = Cairo_StrWidth;
@@ -994,6 +1244,7 @@ configureCairoDevice(pDevDesc dd, CairoDesc *cd, double width, double height, do
   dd->bottom = height;
   dd->top = 0;
 	
+  // Setup font
   fontdesc = getBaseFont(cd);
   pango_font_description_set_size(fontdesc, PANGO_SCALE * ps);
   success = loadFont(fontdesc, cd);
@@ -1006,6 +1257,8 @@ configureCairoDevice(pDevDesc dd, CairoDesc *cd, double width, double height, do
   pango_context_set_font_description(getPangoContext(cd), fontdesc);
   font_metrics(fontdesc, cd, &cw, &ascent, &descent);
 	
+  pango_font_description_free(fontdesc);
+  
   /* starting parameters */
   ps = 2 * (ps / 2);
   dd->startfont = 1; 
@@ -1015,8 +1268,6 @@ configureCairoDevice(pDevDesc dd, CairoDesc *cd, double width, double height, do
   dd->startlty = LTY_SOLID; 
   dd->startgamma = 1;
 	
-  pango_font_description_free(fontdesc);
-	
   dd->cra[0] = cw;
   dd->cra[1] = ascent + descent;
 	
@@ -1026,8 +1277,8 @@ configureCairoDevice(pDevDesc dd, CairoDesc *cd, double width, double height, do
   dd->yLineBias = 0.1;
 
   /* inches per raster unit */
-  dd->ipr[0] = pixelWidth();
-  dd->ipr[1] = pixelHeight();
+  dd->ipr[0] = 1.0/POINTS_PER_INCH;
+  dd->ipr[1] = 1.0/POINTS_PER_INCH;
 
   /* device capabilities */
   dd->canClip = TRUE;
@@ -1063,7 +1314,7 @@ createCairoDevice(pDevDesc dd, double width, double height, double ps, void *dat
     return FALSE;
   }
 
-  return(configureCairoDevice(dd, cd, width / pixelWidth(), height / pixelHeight(), ps));
+  return(configureCairoDevice(dd, cd, width, height, ps));
 }
 Rboolean
 asCairoDevice(pDevDesc dd, double width, double height, double ps, void *data)
@@ -1074,11 +1325,14 @@ asCairoDevice(pDevDesc dd, double width, double height, double ps, void *data)
 
   if(!(cd = createCairoDesc()))
     return FALSE;
-	
-  if (GTK_IS_WIDGET(data))
-    success = Cairo_OpenEmbedded(dd, cd, GTK_WIDGET(data));
-  else success = Cairo_OpenOffscreen(dd, cd, GDK_DRAWABLE(data));
-  
+
+  if (width != -1) { // if width/height specified, we have custom context
+    success = Cairo_OpenCustom(dd, cd, width, height, (cairo_t *)data);
+  } else { // otherwise assume something from GTK
+    if (GTK_IS_WIDGET(data))
+      success = Cairo_OpenEmbedded(dd, cd, GTK_WIDGET(data));
+    else success = Cairo_OpenOffscreen(dd, cd, GDK_DRAWABLE(data));
+  }
   if (!success) {
     freeCairoDesc(dd);
     return FALSE;
@@ -1102,7 +1356,7 @@ initCairoDevice(double width, double height, double ps, void *data, CairoDeviceC
   R_CheckDeviceAvailable();
   BEGIN_SUSPEND_INTERRUPTS {
     /* Allocate and initialize the device driver data */
-    if (!(dev = (pDevDesc) calloc(1, sizeof(NewDevDesc))))
+    if (!(dev = (pDevDesc) calloc(1, sizeof(DevDesc))))
       return NULL;
     if (! init_fun (dev, width, height, ps, data)) {
       free(dev);
@@ -1119,6 +1373,8 @@ initCairoDevice(double width, double height, double ps, void *data, CairoDeviceC
 }
 
 
+
+// Width and Height are in inches
 void
 do_Cairo(double *in_width, double *in_height, double *in_pointsize, char **surface)
 {
@@ -1134,7 +1390,8 @@ do_Cairo(double *in_width, double *in_height, double *in_pointsize, char **surfa
   }
   ps = *in_pointsize;
  
-  initCairoDevice(width, height, ps, surface, (CairoDeviceCreateFun)createCairoDevice);
+  initCairoDevice(width*POINTS_PER_INCH, height*POINTS_PER_INCH, 
+                  ps, surface, (CairoDeviceCreateFun)createCairoDevice);
 
   vmaxset(vmax);
   /*   return R_NilValue; */
@@ -1142,14 +1399,11 @@ do_Cairo(double *in_width, double *in_height, double *in_pointsize, char **surfa
 
 
 SEXP
-do_asCairoDevice(SEXP widget, SEXP pointsize)
+do_asCairoDevice(SEXP widget, SEXP pointsize, SEXP width, SEXP height)
 {
-  void *drawing = R_ExternalPtrAddr(widget);
-  double ps;
-  SEXP ans = Rf_allocVector(LGLSXP, 1);
-
-  ps = REAL(pointsize)[0];
-  LOGICAL(ans)[0] = (initCairoDevice(-1, -1, ps, drawing, asCairoDevice) != NULL);
-
-  return(ans);
+  pDevDesc pd = initCairoDevice(asReal(width), asReal(height),
+                                asReal(pointsize),
+                                R_ExternalPtrAddr(widget),
+                                asCairoDevice);
+  return ScalarLogical(pd != NULL);
 }
